#include <stdio.h>
#include <stdlib.h>
#include <upc.h>
//
// A is the data structure used in this program; A_serial was used
// in the serial program
// A_serial[i] = A[i%THREADS][i/THREADS]
//                                           i = 0..n-1
// A_serial[MYTHREAD+i_local*THREADS] = A[MYTHREAD][i_local]
//                                           i_local = 0..rows_per_thread-1
//                                           MYTHREAD = 0..THREADS-1 
typedef struct {
  shared [] float *row;
  size_t row_size;
} Tri_matrix;

shared [1] size_t n, i_row;
shared [ ] Tri_matrix * shared[1] A[THREADS];

int main(void) {
  size_t i, i_local, j, rows_per_thread, n_elem;
  int next_thread;
  float *row;

  if (MYTHREAD == 0) {
    printf("number of rows (1..100):\n");
    scanf("%li",(size_t *) &n);
    printf("which row should be printed (1..%1li):\n",n);
    scanf("%li",(size_t *) &i_row);
//   casts to (size_t *) are required and work because data are located
//   on executing thread
  }
  upc_barrier;
// n and i_row are now accessible from each thread
  rows_per_thread = n / THREADS;
  if (n%THREADS > 0) rows_per_thread++;

  A[MYTHREAD] = (shared [ ] Tri_matrix *) 
                 upc_alloc(rows_per_thread * sizeof(Tri_matrix));
//
// create triangular matrix
  i_local = 0;
  n_elem = 0;
  for (i=MYTHREAD; i<n; i+=THREADS) {
    A[MYTHREAD][i_local].row_size = n - i;
    A[MYTHREAD][i_local].row = 
           (shared [] float *) upc_alloc(A[MYTHREAD][i_local].row_size*sizeof(float));
    i_local++;
    n_elem += n - i;
  }
//
// initialize matrix A(i, j) = i + j
  i_local = 0;
  for (i=MYTHREAD; i<n; i+=THREADS) {
//  cast to local pointer for performance
    row = (float *) A[MYTHREAD][i_local].row;
    for (j=0; j<n-i; j++) {
      row[j] = (float) (i + 1) + (float) (j + 1);
    }
    i_local++;
  }
//
// write (one-based) row number i_row to standard output
  if ( (i_row - 1)%THREADS == MYTHREAD && i_row <= n ) {
    printf("[me=%2i]  Row %li on thread %i:",  MYTHREAD, i_row, (i_row - 1)%THREADS);
    i_local = (i_row - 1)/THREADS;
    for (j=0; j<n-i_row+1; j++) {
      printf(" %4.0f", A[MYTHREAD][i_local].row[j]);
    }
    printf("\n");
  }
  printf("Number of elements on thread %i: %li\n", 
          MYTHREAD,n_elem);
//
// clean up
  upc_barrier; 
  i_local = 0;
  for (i=MYTHREAD; i<n; i+=THREADS) {
    upc_free(A[MYTHREAD][i_local].row);
    i_local++;
  }
  upc_barrier;  // should not be necessary
  upc_free(A[MYTHREAD]);
}
